import unittest
import StringIO

from sdetools.analysis_integration.mapping import Mapping, MappingError
from sdetools.sdelib.commons import Error

class TestMapping(unittest.TestCase):

    def setUp(self):
        pass

    def test_load_empty_csv(self):
        csv_file = StringIO.StringIO()
        csv_file.write('%s\n' % ','.join(Mapping.EXPECTED_CSV_HEADERS))
        csv_file.seek(0)

        mapping = Mapping()
        mapping._load_mapping_from_csv(csv_file)
        self.assertTrue(mapping.count() == 0, 'Expected no task->weakness mappings')

        csv_file.close()

    def test_load_missing_headers_csv(self):
        csv_file = ("A,B,C,D\n"
                    "T1,Title1,weakness_id1,weakness_title 1\n"
                   )

        mapping = Mapping()
        self.assert_exception(MappingError, "FATAL ERROR: Missing headers: Task ID, Task Title, Weakness ID, Weakness Title",
                              mapping._load_mapping_from_csv, StringIO.StringIO(csv_file))

    def get_xml_mapping(self):
        xml_file = ('<?xml version="1.0"?>'
                    '<mapping>'
                    '   <task id="T1" title="Title1" confidence="high">'
                    '       <weakness id="w1" title="weakness_title 1"/>'
                    '   </task>'
                    '   <task id="T2" title="Title2" confidence="low">'
                    '       <weakness id="w2" title="weakness_title 2"/>'
                    '       <weakness id="w3" title="weakness_title 3"/>'
                    '   </task>'
                    '   <task id="T1" title="Title1" confidence="high">'
                    '       <weakness id="w4" title="weakness_title 4"/>'
                    '   </task>'
                    '   <task id="T3" title="Title3" confidence="low">'
                    '       <weakness id="w1" title="weakness_title 1"/>'
                    '       <weakness id="*" title="Wildcard"/>'
                    '   </task>'
                    '</mapping>')

        return StringIO.StringIO(xml_file)

    def get_csv_mapping(self):
        csv_file = ("Task ID,Task Title,Weakness ID,Weakness Title,Confidence\n"
                    "T1,Title1,w1,weakness_title 1,high\n"
                    "T2,Title2,w2,weakness_title 2,low\n"
                    "T2,Title2,w3,weakness_title 3,low\n"
                    "T1,Title4,w4,weakness_title 4,high\n"
                    "T3,Title3,w1,weakness_title 1,low\n"
                    "T3,Title3,*,Wildcard,low\n"
                   )
        return StringIO.StringIO(csv_file)

    def test_load_xml_mapping(self):
        mapping = Mapping()
        mapping.load_mapping_from_xml(self.get_xml_mapping())
        self._test_mapping(mapping)

    def test_load_csv_mapping(self):
        mapping = Mapping()
        mapping.load_mapping_from_csv(self.get_csv_mapping())
        self._test_mapping(mapping)

    def _test_mapping(self, mapping):
        self.assertTrue(mapping.count() == 5, 'Expected 4 weakness mappings')

        # check that all tasks are mapped
        mapped_tasks = mapping.get_mapped_tasks()
        self.assert_lists_equal(['T1', 'T2', 'T3'], mapped_tasks, 'Expected 3 mapped tasks')

        # validate the covered weaknesses
        weaknesses = mapping.get_weaknesses()
        self.assert_lists_equal(['w1', 'w2', 'w3', 'w4', '*'], weaknesses, 'Expected 5 weaknesses')

        expected_mapping = {
            'w1': [
                {'id': 'T1', 'confidence': 'high'},
                {'id': 'T3', 'confidence': 'low'},
            ],
            'w2': [
                {'id': 'T2', 'confidence': 'low'},
            ],
            'w3': [
                {'id': 'T2', 'confidence': 'low'},
            ],
            'w4': [
                {'id': 'T1', 'confidence': 'high'},
            ],
            '*': [
                {'id': 'T3', 'confidence': 'low'},
            ]
        }

        # check the task mappings
        for weakness in expected_mapping.keys():
            mapped_tasks = [task['id'] for task in mapping.get_tasks(weakness)]
            expected_tasks = [task['id'] for task in expected_mapping[weakness]]
            self.assert_lists_equal(expected_tasks, mapped_tasks, 'Task mapping for %s is not correct' % weakness)

        # check the task confidences
        for weakness in expected_mapping.keys():
            for expected_task in expected_mapping[weakness]:
                task = mapping.get_task(expected_task['id'])
                self.assertTrue(task['confidence'] == expected_task['confidence'], 
                                'Confidence does not match for %s' % expected_task['id'])

    def test_task_matches_on_wildcard(self):
        mapping = Mapping()
        mapping.load_mapping_from_csv(self.get_csv_mapping())
        tasks = mapping.get_tasks("unexpected_weakness")
        self.assertTrue(len(tasks)==1, 'Expected 1 task to match for an unexpected weakness')
        self.assert_lists_equal(['T3'], [tasks[0]['id']], 'Unmatched task mapped for wildcard mapping')
        
    def assert_lists_equal(self, list1, list2, assert_message):
        self.assertTrue(len(list1) == len(list2) and sorted(list1) == sorted(list2), assert_message)

    def assert_exception(self, exception, reason, fn, *args):
        if not exception:
            raise AssertionError('No exception type specified')
        if not fn:
            raise AssertionError('No function specified')

        try:
            fn.__call__(*args)
            raise AssertionError('Expected an exception to be thrown')
        except Error, err:
            self.assertEqual(type(err), exception, 'Expected exception type %s, Got %s' % (exception, type(err)))
            if reason:
                self.assertTrue(reason in str(err), "Expected error message '%s', Got '%s'" % (reason, str(err)))

