import operator

NON_FIELD_ERRORS = '__all__'


class ProjectConnectorException(Exception):
    """
    An error while fetching project data.

    Expects either a string or a list of string of error messages for
    general errors. A dict containing a key, value pair of field names, and
    a list of error messages can be passed for field specific errors.
    Alternatively NON_FIELD_ERRORS key can also be used for general errors

    E.g
    ProjectConnectorException("General error message")

    ProjectConnectorException([
        "General error message 1",
        "General error message 2"
    ])

    ProjectConnectorException({
        'field_1': ['field error message 1'],
        'field_2': ['field error message 2'],
        NON_FIELD_ERRORS: ['general error message']
    })

    NOTE: This class is based on the code from Django's ValidationError
    """

    def __init__(self, message, code=None, params=None):
        if isinstance(message, dict):
            self.error_dict = message
        elif isinstance(message, list):
            self.error_list = message
        else:
            self.code = code
            self.params = params
            self.message = message
            self.error_list = [self]

    @property
    def message_dict(self):
        message_dict = {}
        for field, messages in self.error_dict.items():
            message_dict[field] = []
            for message in messages:
                if isinstance(message, ProjectConnectorException):
                    message_dict[field].extend(message.messages)
                else:
                    message_dict[field].append(message)
        return message_dict

    @property
    def messages(self):
        if hasattr(self, 'error_dict'):
            message_list = reduce(operator.add, self.error_dict.values())
        else:
            message_list = self.error_list

        messages = []
        for message in message_list:
            if isinstance(message, ProjectConnectorException):
                params = message.params
                message = message.message
                if params:
                    message %= params
            messages.append(message)
        return messages

    def __str__(self):
        if hasattr(self, 'error_dict'):
            return repr(self.message_dict)
        return repr(self.messages)

    def __repr__(self):
        return 'ProjectConnectorException(%s)' % self

    def update_error_dict(self, error_dict):
        if hasattr(self, 'error_dict'):
            if error_dict:
                for k, v in self.error_dict.items():
                    error_dict.setdefault(k, []).extend(v)
            else:
                error_dict = self.error_dict
        else:
            error_dict[NON_FIELD_ERRORS] = self.error_list
        return error_dict