import imp
import os


class ProjectConnectorPluginManager(object):
    """
    Manager class for Project Connector Plugins.

    This class is used to register and fetch plugins.
    """

    _plugins = None

    @classmethod
    def register_plugin(cls, plugin_cls):
        """
        Registers a plugin and adds its instance to the plugin registry.
        """
        if cls._plugins is not None:
            cls._plugins.append(plugin_cls())

    @classmethod
    def _get_plugin_modules(cls, plugins_folder):
        """
        Returns a list of plugin modules.

        Scans the plugin folder and returns a list of plugin modules.
        """
        plugin_modules = []

        # walk the plugins folder
        for root, subdir, files in os.walk(plugins_folder):

            # check all sub directories
            for directory in subdir:

                directory_path = os.path.join(root, directory)
                directory_list = os.listdir(directory_path)

                for filename in directory_list:
                    modname, ext = os.path.splitext(filename)

                    if ext == ".py":
                        filename, path, desc = imp.find_module(modname, [directory_path])
                        plugin_modules.append((modname, filename, path, desc,))

        return plugin_modules

    @classmethod
    def get_plugins(cls, plugins_folder=None):
        """
        Returns a list of plugin instances.

        Loads the plugins and returns a list of plugin instances.
        """
        plugin_modules = cls._get_plugin_modules(plugins_folder)

        # load plugins again only if a plugin was added or removed
        if (cls._plugins is not None and
            len(plugin_modules) == len(cls._plugins)):
            return cls._plugins

        cls._plugins = []

        for plugin_module in plugin_modules:
            modname, filename, path, desc = plugin_module

            if filename:
                # loading the module auto-registers it
                imp.load_module(modname, filename, path, desc)

        return cls._plugins

    @classmethod
    def get_plugin(cls, plugins_folder, plugin_id):
        """Returns a plugin with the given plugin id"""
        plugins = cls.get_plugins(plugins_folder)
        plugin = next((p for p in plugins if p.id == plugin_id), None)

        if not plugin:
            raise AssertionError("No plugin found with id: {0}".format(plugin_id))

        return plugin
