from abc import ABCMeta, abstractmethod, abstractproperty

from sdetools.connector_plugin.manager import ProjectConnectorPluginManager


class ProjectConnectorPluginMeta(ABCMeta):
    """
    Meta class for the Project Connector Plugin class.

    Auto-registers the plugin in the manager class.
    """

    def __init__(cls, name, *args, **kwargs):
        if name != 'BaseProjectConnectorPlugin':
            ProjectConnectorPluginManager.register_plugin(cls)


class BaseProjectConnectorPlugin(object):
    """
    Project connector Plugin base class.

    Project connector plugins should derive from this class and implement the
    abstract methods and properties.
    """

    __metaclass__ = ProjectConnectorPluginMeta

    @abstractproperty
    def id(self):
        """
        Unique id of the connector.
        """

    @abstractproperty
    def name(self):
        """
        Name of the connector.
        """
        pass

    @abstractproperty
    def description(self):
        """
        Description of the connector.
        """
        pass

    @abstractmethod
    def configuration_params(self):
        """
        Parameters for configuring the plugin.

        A list of ConnectorParam objects that describe the configuration
        attributes for e.g. username, token, url, etc.
        """
        pass

    @abstractmethod
    def lookup_params(self):
        """
        Parameters for looking up a project in the external system(s).

        A list of ConnectorParam objects that describe the lookup parameters,
        used to lookup a project in an external system(s).
        """
        pass

    @abstractmethod
    def expected_custom_attributes(self):
        """
        Custom attributes that the plugin expects to be set in SDE.

        A list of CustomAttribute objects that describe the attributes.
        """
        pass

    @abstractmethod
    def read_only_attributes(self):
        """
        A list of strings representing attributes that SDE should
        not override.
        """
        pass

    @abstractmethod
    def fetch_project_data(self, configuration, lookup_data):
        """
        Returns the data for the project associated with the given id.

        Args:

            id:            Id of the project in the external system.
            configuration: A dictionary containing key value pairs of
                           configuration param names and their values.
            lookup_data:   A dictionary containing key value pairs of
                           lookup attributes and their values.

        Returns:
            Expected to return a dictionary containing key value pairs of
            project attributes and their values. 'name' is a required
            attribute. Custom attributes should be nested under a key called
            'custom_attributes'. In case of an error, the method is expected
            to return a dictionary containing a key called 'errors' with a
            value of a list of error messages.
        """
        pass
