import datetime
import logging
import StringIO
import time

from sdetools.sdelib.commons import UsageError
from sdetools.extlib.suds.client import Client
from sdetools.analysis_integration.base_integrator import BaseIntegrator, IntegrationError
from sdetools.modules.import_checkmarx.checkmarx_xml_importer import CheckmarxXMLImporter
from sdetools.modules.import_checkmarx.checkmarx_zip_importer import CheckmarxZIPImporter

logging.getLogger('sdetools.extlib.suds').setLevel(logging.WARNING)

__all__ = ['CheckmarxIntegrator']

# Number of seconds to wait for report status
REPORT_CHECK_DELAY_SECONDS = 4


def cmp_scan_display_data(x, y):
    """
    Implement a comparison function for a Checkmarx ScanDisplayData entry.

    There are two datetime values in this dictionary.

    QueuedDateTime: The date and time when a scan was started
    FinishedDateTime: The date and time when a scan was completed. May not exist if the scan is in progress
    """

    x_datetime = None
    y_datetime = None

    if 'FinishedDateTime' in x:
        x_datetime = x['FinishedDateTime']

    if 'FinishedDateTime' in y:
        y_datetime = y['FinishedDateTime']

    if not x_datetime:
        return -1
    elif not y_datetime:
        return 1

    x = datetime.datetime(year=x_datetime['Year'],
                          month=x_datetime['Month'],
                          day=x_datetime['Day'],
                          hour=x_datetime['Hour'],
                          minute=x_datetime['Minute'],
                          second=x_datetime['Second'])

    y = datetime.datetime(year=y_datetime['Year'],
                          month=y_datetime['Month'],
                          day=y_datetime['Day'],
                          hour=y_datetime['Hour'],
                          minute=y_datetime['Minute'],
                          second=y_datetime['Second'])

    if x < y:
        return -1
    elif x > y:
        return 1
    else:
        return 0


class CheckmarxSOAPApi(object):
    def __init__(self, config):
        super(CheckmarxSOAPApi, self).__init__()
        self.config = config
        self.session_id = None
        self.checkmarx_wsclient = None
        config.opts.add('analysis_method', 'Checkmarx server method', default='')
        config.opts.add('analysis_server', 'Checkmarx server', default='')
        config.opts.add('analysis_user', 'Checkmarx user name', default='')
        config.opts.add('analysis_pass', 'Checkmarx password', default='')

    @property
    def web_service_url(self):
        if hasattr(self, '_web_service_url'):
            return self._web_service_url

        url = "%s://%s/Cxwebinterface/CxWsResolver.asmx?WSDL" % (
            self.config['analysis_method'], self.config['analysis_server']
        )
        resolver = Client(url)
        response = resolver.service.GetWebServiceUrl(APIVersion=1, ClientType="SDK")
        if not response.IsSuccesfull:
            raise CheckmarxIntegrationError("Failed to resolve API url: %s" % response.ErrorMessage)

        self._web_service_url = response.ServiceURL+"?WSDL"

        return self._web_service_url

    def login(self):
        """
        Login to Checkmarx server and capture a session identifier
        """
        self.checkmarx_wsclient = Client(self.web_service_url)
        credentials = {"User": self.config['analysis_user'], "Pass": self.config['analysis_pass']}
        # lcid 1033 is English
        response = self.checkmarx_wsclient.service.Login(applicationCredentials=credentials, lcid=1033)
        if not response.IsSuccesfull:
            raise CheckmarxIntegrationError("Login unsuccessful: %s" % response.ErrorMessage)

        self.session_id = response.SessionId

    def get_projects(self):
        """
        Retrieve the list of scan projects
        """
        response = self.checkmarx_wsclient.service.GetProjectsDisplayData(self.session_id)
        if not response.IsSuccesfull:
            raise CheckmarxIntegrationError("Failed to retrieve list of projects: %s" % response.ErrorMessage)

        return response.projectList

    def get_scans(self):
        """
        Retrieve the long list of scans, including any in-progress
        """
        response = self.checkmarx_wsclient.service.GetScansDisplayDataForAllProjects(self.session_id)
        if not response.IsSuccesfull:
            raise CheckmarxIntegrationError("Failed to retrieve scans: %s" % response.ErrorMessage)

        return response.ScanList

    def generate_report(self, scan_id):
        """
        Initiate the generation of an XML report containing the results of a scan
        """
        scan_parameters = {"ScanID": scan_id, "Type": "XML"}
        response = self.checkmarx_wsclient.service.CreateScanReport(self.session_id, scan_parameters)
        if not response.IsSuccesfull:
            raise CheckmarxIntegrationError("Failed to generate report: %s" % response.ErrorMessage)

        return response.ID

    def get_report_status(self, report_id):
        """
        Find out if a prior request to generate a report has completed
        """
        response = self.checkmarx_wsclient.service.GetScanReportStatus(self.session_id, report_id)
        if not response.IsSuccesfull:
            raise CheckmarxIntegrationError("Failed to query report status: %s" % response.ErrorMessage)

        return response

    def get_report(self, report_id):
        """
        Retrieve a report given a report_id
        """
        response = self.checkmarx_wsclient.service.GetScanReport(self.session_id, report_id)
        if not response.IsSuccesfull:
            raise CheckmarxIntegrationError("Failed to retrieve report: %s" % response.ErrorMessage)

        return response


class CheckmarxIntegrationError(IntegrationError):
    pass


class CheckmarxIntegrator(BaseIntegrator):
    TOOL_NAME = "checkmarx"
    DEFAULT_MAPPING_FILE = 'sde_checkmarx_map.csv'

    def __init__(self, cx_api, config):
        self.cx_api = cx_api

        supported_input_types = {
            'file': ['xml', 'zip'],
            'network': 'https'
        }

        config.opts.add('integration_mode', "Integration mode: (api or file)", default='file')
        config.opts.add('project_name', 'Checkmarx Project name', default='')
        config.opts.add('analysis_report_timeout', 'Maximum number of seconds to wait for report generation', default='240')

        super(CheckmarxIntegrator, self).__init__(config, self.TOOL_NAME, supported_input_types)

    def initialize(self):
        self.config['integration_mode'] = self.config['integration_mode'].lower()

        if self.config['integration_mode'] not in ['api', 'file']:
            raise UsageError("Invalid value for integration_mode. Valid values are: api or file")

        if self.config['integration_mode'] == 'api':
            for config_key in ['analysis_method', 'analysis_server', 'analysis_user', 'analysis_pass', 'project_name']:
                if not self.config[config_key]:
                    raise UsageError("Missing value for option %s" % config_key)

            # disable file support
            self.supported_input_types.pop('file')

        super(CheckmarxIntegrator, self).initialize()

        try:
            self.config['analysis_report_timeout'] = int(self.config['analysis_report_timeout'])
        except (ValueError, TypeError):
            raise UsageError('Invalid value for analysis_report_timeout: %s' % self.config['analysis_report_timeout'])

    def parse_report_file(self, report_file, report_type):
        if report_type == 'xml':
            importer = CheckmarxXMLImporter()
        elif report_type == 'zip':
            importer = CheckmarxZIPImporter()
        elif report_type == 'xml_string':
            importer = CheckmarxXMLImporter()
            report = report_file
            report_file = StringIO.StringIO()
            report_file.write(report)
        else:
            raise UsageError("Unsupported file type (%s)" % report_type)

        importer.parse(report_file)

        self.findings = importer.findings
        self.report_id = importer.id
        return importer.findings, importer.id

    def parse(self):
        if self.config['integration_mode'] == 'file':
            return super(CheckmarxIntegrator, self).parse()
        elif self.config['integration_mode'] == 'api':
            report_xml_dom = self.retrieve_report()
            self.findings, self.report_id = self.parse_report_file(report_xml_dom, 'xml_string')
        else:
            raise CheckmarxIntegrationError("Invalid value for 'integration_mode': %s" % self.config['integration_mode'])

    def _make_finding(self, item):
        return {'weakness_id': item['id'], 'description': item['id'], 'type': 'check'}

    def generate_findings(self):
        return [self._make_finding(item) for item in self.findings]

    def retrieve_latest_scan_id(self):
        scans = self.cx_api.get_scans()

        # Sort the scans by the date and time they were finished, in descending order
        sorted_scans = sorted(scans['ScanDisplayData'], cmp=cmp_scan_display_data, reverse=True)

        for scan in sorted_scans:
            if scan['ProjectName'] == self.config['project_name'] and not scan['IsIncremental']:
                return scan['ScanID']

        raise CheckmarxIntegrationError("Failed to retrieve latest full scan for %s" % self.config['project_name'])

    def retrieve_report(self):
        """
        Retrieve the latest XML project report using the Checkmarx API
        """
        self.cx_api.login()
        scan_id = self.retrieve_latest_scan_id()
        report_id = self.cx_api.generate_report(scan_id)
        report_status = self.cx_api.get_report_status(report_id)

        # Set a timeout for when to stop waiting for a report to get generated
        timeout = datetime.datetime.now() + datetime.timedelta(seconds=self.config["analysis_report_timeout"])

        while not report_status.IsReady:
            if datetime.datetime.now() > timeout:
                msg = "Report could not be retrieved in less than %s seconds" % self.config["analysis_report_timeout"]
                raise CheckmarxIntegrationError(msg)

            # Give the server some time to generate the report
            time.sleep(REPORT_CHECK_DELAY_SECONDS)

            report_status = self.cx_api.get_report_status(report_id)
            if report_status.IsFailed:
                raise CheckmarxIntegrationError("The server reported that report generation failed")

        # May take some time
        report = self.cx_api.get_report(report_id)

        return report.ScanResults.decode('base64')
