import unittest
import os
import base64

from mock import patch
from sdetools.analysis_integration.tests.base_integration_test import BaseIntegrationTest
from sdetools.modules.import_checkmarx.checkmarx_integrator import CheckmarxIntegrator, CheckmarxSOAPApi, cmp_scan_display_data
from sdetools.sdelib.commons import get_directory_of_current_module
from sdetools.sdelib.testlib.mock_response import MOCK_ALM_RESPONSE, MOCK_SDE_RESPONSE


class TestCheckmarxIntegration(BaseIntegrationTest, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        sde_cmd = "import_checkmarx"
        mapping_file = 'mapping.csv'
        report_file = 'checkmarx.xml'
        integrator = CheckmarxIntegrator

        cls.test_dir = get_directory_of_current_module(cls)
        cls.mock_sde_response = MOCK_SDE_RESPONSE
        cls.mock_alm_response = MOCK_ALM_RESPONSE

        # Patch Suds client for offline tests
        mock_soap = MockSoap
        patch('sdetools.modules.import_checkmarx.checkmarx_integrator.Client', mock_soap).start()

        super(TestCheckmarxIntegration, cls).setUpClass(sde_cmd, mapping_file, report_file, integrator)

    def init_integrator(self):
        cx_api = CheckmarxSOAPApi(self.config)
        self.integrator = self.integrator_cls(cx_api, self.config)
        self.config.import_custom_options()
        self.integrator.config['mapping_file'] = os.path.join(self.test_file_dir, self.mapping_file)
        self.integrator.config['report_file'] = os.path.join(self.test_file_dir, self.report_file)
        self.num_reports = 1

    def expected_number_of_findings(self):
        return 1

    def test_import_findings_assert_unmapped_tasks(self):
        pass

    def test_import_findings_assert_passed_tasks(self):
        self.init_data()
        result = self.integrator.import_findings()

        self.assertEqual(len(result.noflaw_tasks), 2, 'Expected 2 unaffected tasks, got %s' % len(result.noflaw_tasks))
        self.assertTrue('T40' in result.noflaw_tasks, 'Expected to find task 40 in unaffected tasks')
        self.check_analysis_note('T40', 'partial', 'low', 0)
        self.assertTrue('T38' in result.noflaw_tasks, 'Expected to find task 38 in unaffected tasks')
        self.check_analysis_note('T38', 'partial', 'low', 0)

    def test_api_analysis(self):
        self.integrator.config['integration_mode'] = 'api'
        self.integrator.config['project_name'] = 'WebGoat'
        self.integrator.config['analysis_report_timeout'] = 240

        report = self.integrator.retrieve_report()
        report_file = os.path.join(self.test_file_dir, self.report_file)
        report_expected = open(report_file).read()
        self.assertEqual(report, report_expected)

    def test_scan_list_compare(self):
        scan_list = [
            {
                'ScanID': 'Scan1',
                'FinishedDateTime': {
                    'Hour': 5,
                    'Minute': 57,
                    'Second': 53,
                    'Day': 2,
                    'Month': 11,
                    'Year': 2014,
                },
                'IsIncremental': False,
                'ProjectName': "WebGoat",
            },
            {
                'ScanID': 'Scan2',
                'FinishedDateTime': {
                    'Hour': 5,
                    'Minute': 57,
                    'Second': 53,
                    'Day': 2,
                    'Month': 11,
                    'Year': 2013,
                },
                'IsIncremental': False,
                'ProjectName': "WebGoat",
            },
            {
                'ScanID': 'Scan3',
                'FinishedDateTime': {
                    'Hour': 5,
                    'Minute': 57,
                    'Second': 53,
                    'Day': 3,
                    'Month': 11,
                    'Year': 2015,
                },
                'IsIncremental': False,
                'ProjectName': "WebGoat",
            },
            {
                'ScanID': 'Scan5',
                'FinishedDateTime': {
                    'Hour': 5,
                    'Minute': 57,
                    'Second': 53,
                    'Day': 2,
                    'Month': 11,
                    'Year': 2015,
                },
                'IsIncremental': False,
                'ProjectName': "WebGoat",
            },
            {
                'ScanID': 'Scan4',
                'IsIncremental': False,
                'ProjectName': "WebGoat",
            },
        ]

        sorted_scans = sorted(scan_list, cmp=cmp_scan_display_data)
 
        expected_order = {0: 'Scan4', 1: 'Scan2', 2: 'Scan1', 3: 'Scan5', 4: 'Scan3'}

        for scan_entry, scan_id in expected_order.iteritems():
            self.assertEqual(sorted_scans[scan_entry]['ScanID'], scan_id)
 

class MockSoap():
    def __init__(self, wsdlsource):
        self.service = MockService()


class MockService():
    def __init__(self):
        pass

    def GetWebServiceUrl(self, APIVersion, ClientType):
        return type('', (object,), {"IsSuccesfull": True, "ServiceURL": "http://www.example.com"})()

    def Login(self, applicationCredentials, lcid):
        return type('', (object,), {"IsSuccesfull": True, "SessionId": ""})()

    def GetProjectsDisplayData(self, session_id):
        return type('', (object,), {"IsSuccesfull": True, "projectList": ""})()

    def GetScansDisplayDataForAllProjects(self, session_id):
        return type('', (object,), {
                             "IsSuccesfull": True,
                             "ScanList": {
                                 'ScanDisplayData': [{
                                     'ScanID': 1000714,
                                     'FinishedDateTime': {
                                         'Hour': 5,
                                         'Minute': 57,
                                         'Second': 53,
                                         'Day': 2,
                                         'Month': 11,
                                         'Year': 2015,
                                     },
                                     'IsIncremental': False,
                                     'ProjectName': "WebGoat",
                                 }],
                             }
                         })()

    def CreateScanReport(self, session_id, scan_parameters):
        return type('', (object,), {"IsSuccesfull": True, "ID": ""})()

    def GetScanReportStatus(self, session_id, report_id):
        return type('', (object,), {"IsSuccesfull": True, "IsReady": True, "IsFailed": False})()

    def GetScanReport(self, session_id, report_id):
        # Read in checkmarx sample report file
        results = open(os.path.join(get_directory_of_current_module(self), "files/checkmarx.xml")).read()
        encoded_results = base64.b64encode(results) # Encode results in base64 as checkmarx does
        return type('', (object,), {"IsSuccesfull": True, "ScanResults": encoded_results})()
