#!/usr/bin/python
import csv
import sys
import os
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))

from xml.dom import minidom

class MappingError(Exception):
    pass


class Mapping:
    def __init__(self):
        self.mapping = {}
        self.cwe_checks = {}
        self.base_tasks = {}
        self.checks = []
        self.task_map = {'186':[], '193':[]}

        # these are good candidates for CWE mapping
        self.appcheck_group = ['5', '6', '12', '30', '60', '586', '587', '657', '703', '1202', '2039']
        # these should be considered 3rd party - T186
        self.thirdparty_group = ['589', '594', '606']
        self.thirdparty_checks = []

    def load_other_cwe_mapping(self, other_mapping_file):
        with open(other_mapping_file, 'rb') as csv_file:
            reader = csv.DictReader(csv_file)
            self.other_cwe_mapping = {}
            for row in reader:
                if not row['Query'] in self.other_cwe_mapping:
                    self.other_cwe_mapping[row['Query']] = row['CWE-ID']
                else:
                    continue

    def load_base_mapping_from_xml(self, mapping_file):
        try:
            base = minidom.parse(mapping_file)
        except Exception, e:
            raise MappingError("An error occurred opening mapping file '%s': %s" % (mapping_file, e))

        cwe_mapping = {}
        for task in base.getElementsByTagName('task'):
            self.base_tasks[task.attributes['id'].value ] = task
            for cwe in task.getElementsByTagName('weakness'):
                tasks = []
                if cwe_mapping.has_key(cwe.attributes['id'].value):
                    tasks = cwe_mapping[cwe.attributes['id'].value]
                tasks.append(task.attributes['id'].value)
                cwe_mapping[cwe.attributes['id'].value] = tasks

        self.mapping = cwe_mapping

    def remap(self, unmapped_cwe_file_name):
        fp = open(unmapped_cwe_file_name, "w")
        fp.write("Package,Check,CWE-ID,Status\n")

        self.task_map['186'].extend(self.thirdparty_checks)
        self.task_map['193'].extend([{'check_id': '*', 'package':'', 'check_name': 'Unmapped Check', 'cwe_id':'0'}])

        keys = sorted(self.cwe_checks.iterkeys())
        for cwe in keys:
            if self.mapping.has_key(cwe):
                tasks = self.mapping[cwe]
                for task in tasks:
                    task_checks = []
                    if task in self.task_map.keys():
                        task_checks = self.task_map[task]

                    for check in self.cwe_checks[cwe]:

                        # If the check is assigned here already skip it
                        if self.check_mapped(self.task_map, check['check_id'], task):
                            continue

                        # If the check is assigned elsewhere do not put it in the catch-all task
                        if task == '193' and self.check_mapped(self.task_map, check['check_id'], task):
                            continue

                        task_checks.append(check)

                    self.task_map[task] = task_checks
            else:  # map to 193
                    task_mapping_found = False
                    task_checks = []
                    task = '193'
                    if task in self.task_map.keys():
                        task_checks = self.task_map[task]
                    for check in self.cwe_checks[cwe]:

                        fp.write("%s,%s,%s,unmapped\n" % (check['package'],check['check_id'],check['cwe_id']))

                        # If the check is assigned elsewhere do not put it in the catch-all task
                        if self.check_mapped(self.task_map, check['check_id']):
                            continue

                        for chk in task_checks:
                            if chk['check_id'] == check['check_id']:
                                task_mapping_found = True
                        if not task_mapping_found:
                            task_checks.append(check)
                    self.task_map[task] = task_checks

        fp.close()

    def load_weaknesses_from_xml(self, weakness_file):
        with open(weakness_file, 'rb') as csv_file:
            reader = csv.DictReader(csv_file)
            cwe_checks = {}
            for row in reader:

                check = {}
                check['check_id'] = row['Query']

                # Keep a running tally of the checks
                if not check['check_id'] in self.checks:
                    self.checks.append(check['check_id'])
                else:
                    continue

                # try to map immediately based on check grouping, without CWE
                if self.check_mapped(self.task_map, check['check_id']):
                    continue
                elif self.check_in_list(self.thirdparty_checks, check['check_id']):
                    continue

                check['cwe_id'] = row['CWEID']
                check['package'] = row['Package']

                if int(check['cwe_id']) > 10000 and check['check_id'] in self.other_cwe_mapping:
                    check['cwe_id'] = self.other_cwe_mapping[check['check_id']]

                weaknesses = []
                if cwe_checks.has_key(check['cwe_id']):
                    weaknesses = cwe_checks[check['cwe_id']]
                weaknesses.append(check)
                cwe_checks[check['cwe_id']] = weaknesses

            self.cwe_checks = cwe_checks

    def output_mapping(self, file_name):
        fp = open(file_name, "w")
        keys = sorted(self.task_map.iterkeys())
        fp.write('<mapping weaknesses="%d">\n' % len(self.checks))
        for task_id in keys:
            task_checks = self.task_map[task_id]
            fp.write('\t<task id="T%s" title="%s" confidence="%s">\n' % (
                task_id, self.base_tasks[task_id].attributes['title'].value, "low"))
            task_checks = sorted(task_checks, key=lambda x: x['check_id'])
            for weakness in task_checks:
                fp.write('\t\t<weakness id="%s" type="check" title="%s" package="%s" cwe="%s" />\n' % 
                         (weakness['check_id'], weakness['check_id'], weakness['package'], weakness['cwe_id']))
            fp.write('\t</task>\n')
        fp.write('</mapping>\n')
        fp.close()

    def check_in_list(self, a_list, check_id):
        for item in a_list:
            if item['check_id'] == check_id:
                return True
        return False

    def check_mapped(self, task_map, check_id, selected_task=None):
        keys = task_map.iterkeys()

        if self.check_in_list(self.thirdparty_checks, check_id):
            return True

        for task_id in keys:
            if selected_task:
                if task_id != selected_task:
                    continue
            task_checks = task_map[task_id]
            if self.check_in_list(task_checks, check_id):
                return True

        return False

    def find_missing_checks(self):
        missing_checks = False
        for check_id in self.checks:
            if not self.check_mapped(self.task_map, check_id):
                print "*** Check %s is not mapped" % check_id
                missing_checks = True
        return missing_checks

def main(argv):

    usage = "Usage: %s <base mapping> <custom map> <webinspect groups> <webinspect checks> <unmapped cwes> <webinspect map> \n" % (argv[0])

    if len(argv) == 2 and argv[1] == 'help':
        print usage
        print "base mapping: (input) docs/base_cwe/sde_cwe_map.xml"
        print "Checkmarx checks: (input) Checkmarx queries in csv format [i.e. checkmarx_checks.csv]"
        print "Checkmarx-CWE map: (input) Checkmarx Queries mapped to CWE"
        print "unmapped cwes: (output) list of cwes that were not mapped using the mapping"
        print "Checkmarx map: (output) Checkmarx map file"
        sys.exit(0)
    elif len(argv) < 5:
        print usage
        sys.exit(1)

    base = Mapping()

    print "Loading CWE map..."
    base.load_base_mapping_from_xml(argv[1])

    print "Load other Checkmarx-CWE mapping..."
    base.load_other_cwe_mapping(argv[3])

    print "Loading Checkmarx Checks..."
    base.load_weaknesses_from_xml(argv[2])

    print "Re-mapping Checks... unmapped CWE's recorded in %s" % argv[4]
    base.remap(argv[4])

    print "Validating..."
    if base.find_missing_checks():
        print "Aborted"
        sys.exit(1)

    print "Outputting Checkmarx map to %s..." % argv[5]
    base.output_mapping(argv[5])

    print "Done."        

if __name__ == "__main__":
    main(sys.argv)

