from sdetools.extlib.defusedxml import minidom
from sdetools.sdelib.commons import UsageError
from sdetools.analysis_integration.base_integrator import BaseIntegrator, IntegrationError
from sdetools.sdelib.restclient import RESTBase, APIError

from sdetools.sdelib import log_mgr
logger = log_mgr.mods.add_mod(__name__)

__all__ = ['VeracodeIntegrator']

REQUIRED_ATTRIBS = ['issueid', 'cweid', 'categoryid', 'categoryname', 'description', 'severity', 'module', 'remediation_status']
LOCATION_ATTRIBS = ['sourcefilepath', 'sourcefile', 'line', 'location']


class VeracodeAPI(RESTBase):

    def __init__(self, config):
        super(VeracodeAPI, self).__init__('analysis', 'Veracode', config)

    def parse_response(self, result, headers):
        try:
            return minidom.parseString(result)
        except Exception:
            raise VeracodeIntegrationError('Unable to process XML data: %s' % str(result)[:200])


class VeracodeIntegrationError(IntegrationError):
    pass


class VeracodeIntegrator(BaseIntegrator): 
    TOOL_NAME = "veracode"
    DEFAULT_MAPPING_FILE = 'sde_veracode_map.csv'

    def __init__(self, vc_api, config):
        self.vc_api = vc_api

        supported_input_types = {
            'file': ['xml'],
            'network': 'https'
        }
        config.opts.add('integration_mode', "Integration mode: (api or file)", default='file')
        config.opts.add('project_name', 'Veracode Project name', default='')

        super(VeracodeIntegrator, self).__init__(config, self.TOOL_NAME, supported_input_types)

    def initialize(self):
        if self.config['integration_mode'] == 'api':
            for config_key in ['analysis_method', 'analysis_server', 'analysis_user', 'analysis_pass', 'project_name']:
                if not self.config[config_key]:
                    raise UsageError("Missing value for option %s" % config_key)

            # disable file support
            self.supported_input_types.pop('file')
            super(VeracodeIntegrator, self).initialize()

        elif self.config['integration_mode'] == 'file':
            super(VeracodeIntegrator, self).initialize()
        else:
            raise UsageError("Invalid value for integration_mode. Valid values are: api or file")

    def _make_raw_finding(self, node):
        """
        Extract finding information from an XML node.
        """
        entry = {}
        for attr in REQUIRED_ATTRIBS:
            if attr not in node.attributes.keys():
                raise VeracodeIntegrationError('Required attribute %s missing' % (attr))
            entry[attr] = node.attributes[attr].value
        for attr in LOCATION_ATTRIBS:
            if attr in node.attributes.keys():
                entry[attr] = node.attributes[attr].value
        return entry

    def parse(self):
        if self.config['integration_mode'] == 'file':
            return super(VeracodeIntegrator, self).parse()
        elif self.config['integration_mode'] == 'api':
            report_xml_dom = self.retrieve_report(self.config['project_name'])
            if not report_xml_dom:
                raise VeracodeIntegrationError("Failed to retrieve report for %s" % self.config['project_name'])
            
            self.findings, self.report_id = self.parse_report_file(report_xml_dom, 'xml_dom')

    def parse_report_file(self, report_file, report_type):

        if report_type != 'xml' and self.config['integration_mode'] == 'file':
            raise UsageError("Unsupported file type (%s)" % report_type)

        # Caller sent in an XML document
        if report_type == 'xml_dom':
            base = report_file

        #  XML has to be parsed
        else:
            try:
                base = minidom.parse(report_file)
            except Exception, err:
                raise VeracodeIntegrationError("Error opening report xml (%s): %s" % (report_file, str(err)))

        detailed_reports = base.getElementsByTagName('detailedreport')

        if len(detailed_reports) != 1:
            raise VeracodeIntegrationError('An unexpected number of detailedreport nodes found (%d)' %
                                           len(detailed_reports))
        dr = detailed_reports[0]
        report_id = "%s (%s-b%s)" % (
            dr.attributes['app_name'].value,
            dr.attributes['app_id'].value,
            dr.attributes['build_id'].value
        )
        findings = [self._make_raw_finding(node) for node in base.getElementsByTagName('flaw')]

        # Veracode tracks 'fixed' flaws - prune them out
        for flaw in list(findings):
            if flaw['remediation_status'] == 'Fixed':
                findings.remove(flaw)

        return findings, report_id

    def _make_finding(self, item):
        finding = {'weakness_id': item['cweid'], 'description': item['categoryname']}
        if item.has_key('sourcefilepath'):
            finding['source'] = item['sourcefilepath']
        if item.has_key('line'):
            finding['line'] = item['line']
        if item.has_key('inputvector'):
            finding['source'] = item['inputvector']
        return finding

    def generate_findings(self):
        return [self._make_finding(item) for item in self.findings]

    def retrieve_report(self, app_name):
        logger.info('Retrieving report for %s' % app_name)

        app_id = self.retrieve_app_id_for_name(app_name)
        if not app_id:
            raise VeracodeIntegrationError('Application %s not found' % app_name)
        app_id = int(app_id)

        build_id = self.retrieve_latest_build_id(app_id)
        if not build_id:
            raise VeracodeIntegrationError('No builds found for application: %s' % app_name)
        build_id = int(build_id)

        target_url = 'api/3.0/detailedreport.do?build_id=%d&app_id=%d' % (build_id, app_id)
        try:
            return self.vc_api.call_api(target_url)
        except APIError:
            raise APIError('Failed to retrieve report for application %s, build: %d' % (app_name, build_id))

    def retrieve_app_id_for_name(self, app_name):
        target_url = 'api/4.0/getapplist.do'

        try:
            app_list = self.vc_api.call_api(target_url)
        except APIError, err:
            raise VeracodeIntegrationError ('Unable to retrieve applications: %s' % err)

        for app in app_list.getElementsByTagName('app'):
            if app.attributes['app_name'].value == app_name:
                return app.attributes['app_id'].value

        return None

    def retrieve_latest_build_id(self, app_id):
        target_url = 'api/4.0/getbuildlist.do?app_id=%d' % app_id

        try:
            build_list = self.vc_api.call_api(target_url)
        except APIError, err:
            raise VeracodeIntegrationError ('Unable to retrieve build list: %s' % err)

        # Last eleement is newest build
        return build_list.getElementsByTagName('build')[-1].attributes['build_id'].value 
