# Copyright SDElements Inc
# Extensible two way integration with Mingle

from datetime import datetime
from sdetools.sdelib.commons import json, urlquote_str
from sdetools.sdelib.restclient import RESTBase
from sdetools.sdelib.restclient import URLRequest, APIError
from sdetools.alm_integration.alm_plugin_base import AlmTask, AlmConnector
from sdetools.alm_integration.alm_plugin_base import AlmException
from sdetools.extlib import markdown

from sdetools.sdelib import log_mgr
logger = log_mgr.mods.add_mod(__name__)

TFS_DEFAULT_PRIORITY_MAP = {
    '9-10': '1',
    '7-8': '2',
    '3-6': '3',
    '1-2': '4',
}


class TFSAPIBase(RESTBase):
    def __init__(self, config):
        super(TFSAPIBase, self).__init__('alm', 'TFS', config, '')

    def parse_error(self, result):
        # Try to parse the response as JSON
        try:
            error_response = json.loads(result)
        except ValueError:
            return result

        # Send back all the error messages in one go
        if 'message' in error_response and error_response['message']:
            return error_response['message']
        else:
            return result

class TFSTask(AlmTask):
    """ Representation of a task in TFS"""

    def __init__(self, task_id, work_item, done_statuses):
        self.task_id = task_id
        self.work_item = work_item
        self.done_statuses = done_statuses  # comma-separated list
        self.priority = None
        self.status = None
        self.modified_date = None
        self.revision = work_item['rev']

        for item in work_item['fields']:
            if item == 'System.State':
                self.status = work_item['fields']['System.State']
            elif item == 'System.ChangedDate':
                self.modified_date = work_item['fields']['System.ChangedDate']

    def get_revision(self):
        return self.revision

    def get_task_id(self):
        return self.task_id

    def get_alm_id(self):
        return self.work_item['id']

    def get_status(self):
        """ Translates Mingle status into SDE status """
        if self.status.lower() in [status.lower() for status in self.done_statuses]:
            return 'DONE'
        else:
            return 'TODO'

    def get_timestamp(self):
        """ Returns a datetime object """
        return datetime.strptime(self.modified_date,
                                 '%Y-%m-%dT%H:%M:%S.%fZ')


class TFSConnector(AlmConnector):
    alm_name = 'TFS'
    default_priority_map = TFS_DEFAULT_PRIORITY_MAP

    def __init__(self, config, alm_plugin):
        """ Initializes connection to TFS """
        super(TFSConnector, self).__init__(config, alm_plugin)

        config.opts.add('tfs_iterationpath', 'Iteration Path in TFS',
            default='')
        config.opts.add('tfs_issue_type', 'Issue Type to create in TFS',
            default='User Story')
        config.opts.add('tfs_new_status', 'Status to set for new tasks in TFS',
            default='New')
        config.opts.add('tfs_reopen_status', 'Status to reopen tasks in TFS',
            default='New')
        config.opts.add('tfs_done_statuses', 'Statuses that signify a task is Done in TFS',
            default='Done')
        config.opts.add('tfs_collection', 'Name of collection in TFS',
            default='DefaultCollection')
        config.opts.add('alm_issue_label', 'Tag applied to issue in TFS',
            default='SD-Elements')

        self.project_collections = None
        self.projects = None

    def initialize(self):

        super(TFSConnector, self).initialize()

        for item in ['tfs_done_statuses',
                     'tfs_iterationpath',
                     'tfs_issue_type',
                     'tfs_new_status',
                     'alm_issue_label']:

            if not self.config[item]:
                raise AlmException('Missing %s in configuration' % item)

        self.config.process_list_config('tfs_done_statuses')

        self.mark_down_converter = markdown.Markdown(safe_mode="escape")

    def alm_connect_server(self):
        try:
            self.project_collections = self.alm_plugin.call_api('_apis/projectcollections')
        except APIError, err:
            raise AlmException('Unable to connect to TFS. Please check server URL, ID, password. Reason: %s' % err)

    def alm_connect_project(self):

        # Check if project collection name is valid
        found_collection = False
        for project_collection in self.project_collections['value']:
            if project_collection['name'] == self.config['tfs_collection']:
                found_collection = True
                break

        if not found_collection:
            raise AlmException('Project collection %s is not valid' % self.config['tfs_collection'])

        # Check if project name is valid
        try:
            self.projects = self.alm_plugin.call_api('%s/_apis/projects' % self.config['tfs_collection'])
        except APIError, err:
            raise AlmException('Unable to retrieve list of projects: %s' % err)

        found_project = False
        for project in self.projects['value']:
            if project['name'] == self.config['alm_project']:
                found_project = True
                break

        if not found_project:
            raise AlmException('Project %s could not be found' % self.config['alm_project'])

    def alm_validate_configurations(self):
        pass

    def alm_get_task(self, task):

        task_id = self._extract_task_id(task['id'])

        wiql_query = ('Select [Id] From WorkItems Where [System.IterationPath]=\'%s\' and [System.AreaPath]=\'%s\' '
                      'and [Title] Contains \'%s\'' % (self.config['tfs_iterationpath'], self.config['alm_project'], 
                                                       task['alm_fixed_title'])) 
        data = {
            "query": wiql_query
        }

        target_url = '%s/%s/_apis/wit/wiql?api-version=1.0' % \
                     (self.config['tfs_collection'], urlquote_str(self.config['alm_project']))
        try:
            response = self.alm_plugin.call_api(target_url, method=URLRequest.POST, args=data)
        except APIError, err:
            raise AlmException('Encountered an error searching for task: %s' % err)

        if 'workItems' not in response or not response['workItems']:
            return None

        work_item = response['workItems'][0]

        return self._get_workitem_from_tfs(work_item['id'], task_id)

    def _get_workitem_from_tfs(self, work_item_id, task_id):

        work_item_url = '%s/_apis/wit/workitems?ids=%s' % (self.config['tfs_collection'], work_item_id)
        try:
            response = self.alm_plugin.call_api(work_item_url)
        except APIError, err:
            raise AlmException('Unable to retrieve work item for %s: %s' % (task_id, err))

        if 'count' not in response or not response['count']:
            return None

        return TFSTask(task_id, response['value'][0], self.config['tfs_done_statuses'])

    def alm_add_task(self, task):

        if self.config['alm_issue_label']:
            tags_string = self.config['alm_issue_label']
        
        for tag in task['tags']:
            tags_string += '; ' + tag

        data = [
                {
                    'op': 'add',
                    'path': '/fields/System.Title',
                    'value': task['alm_full_title']
                },
                {
                    'op': 'add',
                    'path': '/fields/System.State',
                    'value': self.config['tfs_new_status']
                },
                {
                    'op': 'add',
                    'path': '/fields/System.IterationPath',
                    'value': self.config['tfs_iterationpath']
                },
                {
                    'op': 'add',
                    'path': '/fields/System.AreaPath',
                    'value': self.config['alm_project']
                },
                {
                    'op': 'add',
                    'path': '/fields/System.WorkItemType',
                    'value': self.config['tfs_issue_type']
                },
                {
                    'op': 'add',
                    'path': '/fields/System.Description',
                    'value': self.sde_get_task_content(task)
                },
                {
                    'op': 'add',
                    'path': '/fields/Microsoft.VSTS.Common.Priority',
                    'value': self.translate_priority(task['priority'])
                },
                {
                    'op': 'add',
                    'path': '/fields/System.Tags',
                    'value': tags_string
                },
            ]

        target_url = '%s/%s/_apis/wit/workitems/$Task?api-version=1.0' % \
                     (self.config['tfs_collection'], urlquote_str(self.config['alm_project']))

        try:
            work_item = self.alm_plugin.call_api(target_url, method=URLRequest.PATCH, args=data, 
                                        call_headers={"content-type": "application/json-patch+json"})
        except APIError, err:
            raise AlmException('Encountered an error creating a new work item: %s' % err)

        if self.config['alm_standard_workflow'] and (task['status']['meaning'] in ['DONE', 'NA']):
            tfs_task = TFSTask(work_item['id'], work_item, self.config['tfs_done_statuses'])
            self.alm_update_task_status(tfs_task, task['status']['meaning'])

        return '%s: %s, URL: %s' % (self.config['tfs_issue_type'], work_item['id'], work_item['_links']['html']['href'])

    def alm_update_task_status(self, alm_task, status):

        if status == 'DONE' or status == 'NA':
            new_state = self.config['tfs_done_statuses'][0]
        elif status == 'TODO':
            new_state = self.config['tfs_reopen_status']

        data = [
                {
                    'op': 'add',
                    'path': '/fields/System.State',
                    'value': new_state
                },
            ]

        work_item_url = '%s/_apis/wit/workitems/%s?api-version=1.0' % (self.config['tfs_collection'], alm_task.get_alm_id())
        try:
            self.alm_plugin.call_api(work_item_url, method=URLRequest.PATCH, args=data, 
                                        call_headers={"content-type": "application/json-patch+json"})
        except APIError, err:
            raise AlmException('Encountered an error updating work item: %s' % err)

        logger.debug('Status changed to %s for %s %s in TFS' %
                     (status, self.config['tfs_issue_type'], alm_task.get_alm_id()))

    def alm_disconnect(self):
        pass

    def convert_markdown_to_alm(self, content, ref):
        return self.mark_down_converter.convert(content)
