# Copyright SDElements Inc
# Extensible two way integration with VersionOne
from datetime import datetime
from sdetools.extlib.defusedxml import minidom
from sdetools.sdelib.commons import json
from sdetools.sdelib.restclient import RESTBase
from sdetools.sdelib.restclient import URLRequest, APIError
from sdetools.alm_integration.alm_plugin_base import AlmTask, AlmConnector
from sdetools.alm_integration.alm_plugin_base import AlmException
from sdetools.extlib import markdown

from sdetools.sdelib import log_mgr
logger = log_mgr.mods.add_mod(__name__)

V1_DEFAULT_PRIORITY_MAP = {
    '7-10': 'High',
    '4-6': 'Medium',
    '1-3': 'Low',
}


class VersionOneAPIBase(RESTBase):
    def __init__(self, config):
        super(VersionOneAPIBase, self).__init__('alm', 'VersionOne', config, '')
        self.set_auth_mode('api_token')

    def post_conf_init(self):
        if self._get_conf('api_token'):
            self.set_auth_mode('api_token')
            self.api_token_header_name = 'Authorization'
            self.config['alm_pass'] = 'Bearer %s' % self._get_conf('api_token')
        else:
            self.set_auth_mode('basic')
        super(VersionOneAPIBase, self).post_conf_init()

    def parse_error(self, result):
        # Try to parse the response as JSON
        try:
            error_response = json.loads(result)
        except ValueError:
            return result

        # Send back all the error messages in one go
        if 'exceptions' in error_response:
            error_messages = []
            for msg in error_response['exceptions']:
                error_messages.append('%s: %s' % (msg['type'], msg['message']))
            return ' '.join(error_messages)
        else:
            return result

    def encode_post_args(self, args, call_headers):
        if 'Content-type' in call_headers and 'json' in call_headers['Content-type']:
            return json.dumps(args)
        return args

    def parse_response(self, result, headers):
        if 'content-type' in headers:
            if  'xml' in headers['content-type']:
                try:
                    return minidom.parseString(result)
                except Exception as e:
                    raise APIError('Unable to process XML data. Reason: %s' % e)
            elif 'json' in headers['content-type']:
                return super(VersionOneAPIBase, self).parse_response(result, headers)
        else:
            return result


class VersionOneTask(AlmTask):
    """ Representation of a task in VersionOne """

    def __init__(self, task_id, asset, done_statuses):
        self.task_id = task_id
        self.asset = asset
        self.done_statuses = done_statuses  # comma-separated list
        self.priority = None
        self.status = asset['status']
        self.modified_date = asset['ChangeDate']

    def get_task_id(self):
        return self.task_id

    def get_alm_id(self):
        return self.asset['_oid']

    def get_status(self):
        """ Translates VersionOne status into SDE status """
        if self.status and self.status.lower() in [status.lower() for status in self.done_statuses]:
            return 'DONE'
        else:
            return 'TODO'

    def get_timestamp(self):
        """ Returns a datetime object """
        return datetime.strptime(self.modified_date[:len(self.modified_date)-1],
                                 '%Y-%m-%dT%H:%M:%S.%f')


class VersionOneConnector(AlmConnector):
    alm_name = 'VersionOne'
    default_priority_map = V1_DEFAULT_PRIORITY_MAP

    def __init__(self, config, alm_plugin):
        """ Initializes connection to VersionOne """
        super(VersionOneConnector, self).__init__(config, alm_plugin)

        config.opts.add('v1_issue_type', 'Issue Type to create in %s' % self.alm_name,
            default='Story')
        config.opts.add('v1_new_status', 'Status to set for new tasks in %s' % self.alm_name,
            default='Future')
        config.opts.add('v1_reopen_status', 'Status to reopen tasks in %s' % self.alm_name,
            default='Future')
        config.opts.add('v1_done_statuses', 'Statuses that signify a task is Done in %s' % self.alm_name,
            default='Done,Accepted')
        config.opts.add('alm_api_token', 'API Token for accessing %s' % self.alm_name, default='')

        self.relations = {}
        self.projects = None

    def initialize(self):
        super(VersionOneConnector, self).initialize()

        for item in ['v1_done_statuses',
                     'v1_issue_type',
                     'v1_new_status',
                     'v1_reopen_status']:

            if not self.config[item]:
                raise AlmException('Missing %s in configuration' % item)

        if self.config['v1_issue_type'] not in ['Story', 'Task']:
            msg = 'Invalid issue type %s specified. Only Story and Task are supported' % self.config['v1_issue_type']
            raise AlmException(msg)

        self.config.process_list_config('v1_done_statuses')
        self.mark_down_converter = markdown.Markdown(safe_mode="escape")

    def alm_connect_server(self):
        """
        Connect to the VersionOne Server

        Information about a set of relations is needed for the integration. This method retrieves
        the relation info related to an asset as well as its field meta
        """
        required_assets = ['Scope', '%sStatus' % self.config['v1_issue_type'], 'WorkitemPriority']

        for asset in required_assets:
            try:
                self.relations[asset] = self.alm_plugin.call_api('rest-1.v1/Data/%s' % asset, 
                                                                 call_headers={'accept': 'application/json'})
            except APIError, err:
                raise AlmException('Unable to connect to server. Please check URL, ID, password. Reason: %s' % err)

        # Retrieve field meta information - V1 returns this only as XML
        try:
            asset_meta = self.alm_plugin.call_api('meta.v1/%s' % self.config['v1_issue_type'],
                                                  call_headers={'accept': 'application/xml'})
        except APIError, err:
            raise AlmException('Unable to retrieve meta from server. Reason: %s' % err)

        self.field_meta = asset_meta.getElementsByTagName('AttributeDefinition')

        # Construct a dictionary of field display names -> field meta since users will be defining custom fields using 
        # their understanding of the field versus the underlying system field name.
        fields = []
        for field in self.field_meta:
            fields.append(field.attributes['displayname'].value)

        # The V1 localization endpoint translates fields into their user-friendly field names in batch
        try:
            field_localization = self.alm_plugin.call_api('loc-2.v1', method=URLRequest.POST, args=fields,
                                                          call_headers={'Content-type': 'application/json'})
        except APIError, err:
            msg = 'Unable to retrieve localization value for %s. Reason: %s' % (field.attributes['name'].value, err)
            raise AlmException(msg)

        # Invert the dictionary so that its easier to lookup fields by their display name
        self.display_names = dict((v,k) for k,v in field_localization.iteritems())

    def _get_display_name_for_field_name(self, field_name):
        """
        Retrieve the display name given the field's system name
        """
        for field in self.field_meta:
            if field.attributes['name'].value == field_name:
                for field_display, field_display_key in self.display_names.iteritems():
                    if field_display_key == field.attributes['displayname'].value:
                        return field_display
        return None

    def _get_field_meta(self, field_name):
        """
        Retrieve field meta information given the user-friendly field display name
        """
        if field_name not in self.display_names:
            return None

        for field in self.field_meta:
            if field.attributes['displayname'].value == self.display_names[field_name]:
                return field

    def _field_needed(self, field_name):
        """
        Return true if field_name is needed by the asset we are trying to create
        """
        field_meta = self._get_field_meta(field_name)
        if field_meta:
            return field_meta.attributes['isrequired'].value.lower() == "true"

        return False

    def _get_relation_instance(self, field_name, relation_value):
        """
        Given an asset field name, retrieve the relation instance that
        corresponds to relation_value.

        Lookup is first done with the Number attribute and then Name.
        """
        field_meta = self._get_field_meta(field_name)
        if not field_meta:
            return None

        # Determine the name of the Asset from field meta
        related_asset = field_meta.getElementsByTagName('RelatedAsset')[0]
        if not related_asset:
            return None

        relation_name = related_asset.attributes['nameref'].value

        # We cache the relation list to minimize duplicate network calls
        if relation_name not in self.relations:
            try:
                self.relations[relation_name] = self.alm_plugin.call_api('rest-1.v1/Data/%s' % relation_name,
                                                                         call_headers={'accept': 'application/json'})
            except APIError, err:
                raise AlmException('Unable to retrieve %s from server. Reason: %s' % (relation_name, err))

        for relation_instance in self.relations[relation_name]['Assets']:
            if ('Number' in relation_instance['Attributes'] and 
                relation_instance['Attributes']['Number']['value'] == relation_value):
                return relation_instance
            elif relation_instance['Attributes']['Name']['value'] == relation_value:
                return relation_instance

        return None

    def alm_validate_configurations(self):
        # Check that we have enough fields defined to go ahead with the sync
        standard_field_names = ['Name', 'Description', 'Scope', 'WorkItemPriority']

        # Collect the system field meta for the custom fields
        custom_field_meta = [self._get_field_meta(field_display_name)
                                   for field_display_name in self.config['alm_custom_fields'].keys()
                                       if self._get_field_meta(field_display_name)]

        # Construct the field names as a list
        new_asset_fields_meta = [
            field_meta.attributes['name'].value for field_meta in custom_field_meta
        ] + standard_field_names

        missing_fields = [field.attributes['name'].value
                            for field in self.field_meta
                                if (field.attributes['isrequired'].value.lower() == "true" and
                                    field.attributes['name'].value not in new_asset_fields_meta)]

        if missing_fields:
            raise AlmException('Missing a value for fields: %s' % ', '.join(missing_fields))

        # Validate the priorities
        available_priorities = [
            priority['Attributes']['Name']['value'] for priority in self.relations['WorkitemPriority']['Assets']
        ]

        pmap = self.config[self.ALM_PRIORITY_MAP]
        missing_priorities = [priority_name
                                 for key, priority_name in pmap.iteritems() if priority_name not in available_priorities]

        if missing_priorities:
            raise AlmException('Incorrect priority mapping values specified: %s' % ', '.join(missing_priorities))

        # Validate the statuses
        relation_status = self.relations['%sStatus' % self.config['v1_issue_type']]
        available_statuses = [
            status['Attributes']['Name']['value'] for status in relation_status['Assets']
        ]

        invalid_statuses = [status for status in self.config['v1_done_statuses'] if status not in available_statuses]

        if invalid_statuses:
            raise AlmException('Invalid v1_done_statuses values: %s. Possible values are: %s' %
                               (', '.join(invalid_statuses), ', '.join(available_statuses)))

        if self.config['v1_new_status'] not in available_statuses:
            raise AlmException('Invalid v1_new_status: %s. Possible values are: %s' %
                               (self.config['v1_new_status'], ', '.join(available_statuses)))

        if self.config['v1_reopen_status'] not in available_statuses:
            raise AlmException('Invalid v1_reopen_status: %s. Possible values are: %s' %
                               (self.config['v1_reopen_status'], ', '.join(available_statuses)))

    def alm_connect_project(self):
        project_field = self._get_display_name_for_field_name('Scope')

        # Check if project is valid
        self.project = self._get_relation_instance(project_field, self.config['alm_project'])
        if not self.project:
            raise AlmException('Project %s could not be found' % self.config['alm_project'])

    def alm_get_task(self, task):
        # Setup the V1-specific query for the task given its title
        query = {
           'from': self.config['v1_issue_type'],
           'select': ['Name', 'ChangeDate', 'Status', 'Number'],
           'where': [{
               'Scope': self.project['id']
            }],
           'find': task['alm_fixed_title'],
           'findin': ['Name']
        }

        try:
            response = self.alm_plugin.call_api('query.v1', method=URLRequest.POST, args=query,
                                                call_headers={'Content-type': 'application/json'})
        except APIError, err:
            raise AlmException('Encountered an error searching for task: %s' % err)

        if not response[0]:
            return None

        # Retrieve the first one we find
        asset = response[0][0]

        # Set a default status value
        asset['status'] = None
        
        # Set the actual status on the asset, if possible, given the status id
        for status in self.relations['%sStatus' % self.config['v1_issue_type']]['Assets']:
            if status['id'] == asset['Status']['_oid']:
                asset['status'] = status['Attributes']['Name']['value']

        return VersionOneTask(asset['_oid'], asset, self.config['v1_done_statuses'])

    def alm_add_task(self, task):
        # Create a dictionary of field->value pairs we are ready to set on the asset
        new_asset_fields = {
            'Name': task['alm_full_title'],
            'Description': self.sde_get_task_content(task),
            'Priority': self.translate_priority(task['priority']),
            'Project': self.config['alm_project'],
            'Status': self.config['v1_new_status'],
        }

        # Update the status to match task
        if (self.config['alm_standard_workflow'] and task['status']['meaning'] in ['DONE', 'NA']):
            new_asset_fields['Status'] = self.config['v1_done_statuses'][0]

        # Merge the fields with whatever the user has selected
        new_asset_fields.update(self.config['alm_custom_fields'])

        # Construct XML describing the attributes we want to set on the asset
        # TODO: Switch to JSON when V1 supports it

        fields_xml = ''
        for field, value in new_asset_fields.iteritems():

            # We are assuming these fields are always applicable to assets
            if field in ['Name', 'Description']:
                fields_xml += '<Attribute name="%s" act="set"><![CDATA[%s]]></Attribute>' % (field, value)
                continue
          
            field_meta = self._get_field_meta(field)

            # Skip the field if the asset does not use it
            if not field_meta:
               continue

            # Skip if the field is readonly
            if field_meta.attributes['isreadonly'].value.lower() == 'true':
                continue

            # If the field is tied to relation, get it using the value and construct the XML
            if field_meta.attributes['attributetype'].value.lower() == 'relation':
                relation_instance = self._get_relation_instance(field, value)
                if not relation_instance:
                    raise AlmException('Could not retrieve relation for field %s with value %s' % (field, value))

                relation_name = field_meta.attributes['name'].value
                fields_xml += '''<Relation name="%s" act="set">
                                        <Asset href="%s" idref="%s"/>
                                 </Relation>''' % (relation_name, relation_instance['href'], relation_instance['id'])

            # The field is a Text, Number, etc.
            else:
                fields_xml += '<Attribute name="%s" act="set"><![CDATA[%s]]></Attribute>' % (field, value)

        xml = '<Asset href="/%s/rest-1.v1/New/%s">%s</Asset>' % (
                  self.alm_plugin._get_context_root(), 
                  self.config['v1_issue_type'],
                  fields_xml) 

        try:
            asset = self.alm_plugin.call_api('rest-1.v1/Data/%s' % self.config['v1_issue_type'],
                                             method=URLRequest.POST, args=xml, 
                                             call_headers={"Content-type": "application/xml",
                                                           "Accept": "application/json"})
        except APIError, err:
            raise AlmException('Encountered an error creating a new asset: %s' % err)

        return 'URL: %s://%s%s/%s.mvc/Summary?oidToken=%s' % (
                   self.config['alm_method'],
                   self.config['alm_server'],
                   self.alm_plugin._get_context_root(),
                   self.config['v1_issue_type'],
                   asset['id'])

    def alm_update_task_status(self, alm_task, status):
        if status in ['DONE', 'NA']:
            new_state = self.config['v1_done_statuses'][0]
        elif status == 'TODO':
            new_state = self.config['v1_reopen_status']

        # Status is a relation that must be retrieve via value
        v1_new_status = self._get_relation_instance('Status', new_state)
        if not v1_new_status: 
             raise AlmException('Status %s not found' % new_state)

        xml = '''<Asset><Relation name="Status" act="set">
			<Asset href="%s" idref="%s" />
		</Relation></Asset>''' % (v1_new_status['href'], v1_new_status['id'])
        
        asset_url = 'rest-1.v1/Data/%s' % alm_task.get_alm_id().replace(':', '/')
        try:
            self.alm_plugin.call_api(asset_url, method=URLRequest.POST, args=xml, 
                                     call_headers={"Content-type": "application/xml", "Accept": "application/json"})
        except APIError, err:
            raise AlmException('Encountered an error updating asset: %s' % err)

        logger.debug('Status changed to %s for %s in %s' % (status, alm_task.get_alm_id(), self.alm_name))

    def alm_disconnect(self):
        pass

    def convert_markdown_to_alm(self, content, ref):
        return self.mark_down_converter.convert(content)
