from abc import abstractproperty


class BaseReportPlugin(object):
    """
    Report Plugin base class.

    Report plugins should derive from this class and implement the
    abstract methods and properties.
    """

    # This dict is populated with meta information that the plugin might
    # need. The following values are available: org, application, project, user
    meta = {}

    @abstractproperty
    def title(self):
        """
        Title of the report.
        """
        pass

    @abstractproperty
    def description(self):
        """
        Description of the report.
        """
        pass

    @abstractproperty
    def ordinal(self):
        """
        Ordinal value of the report. It is used to order the report in the UI.
        """
        pass

    @abstractproperty
    def supported_formats(self):
        """
        A tuple consisting of supported report file formats.
        """
        pass

    @property
    def template(self):
        """
        HTML template for the report.
        """
        return None

    @property
    def filename(self):
        """
        Filename for CSV or PDF formats if supported.
        """
        return None

    @property
    def options(self):
        """
        A list of options for the report.

        These options show up in the UI. Each option must be a dict
        consisting of following:

        name: The programmatic name of the option. The option's value will be
        available by this name on the get_response context object.

        text: The text that shows up in the UI.

        type: Two types are supported - 'boolean' and 'enum'.

        values: Only specific to enum types. Needs to be a list of tuples
        containing the name and value. e.g. ('Phase A', 'phase_a_slug')
        """
        return []

    def get_context(self, request, context):
        """
        A dict containing all the data that is to be made available for
        rendering the report.
        """
        context.update({
            'format': request.GET['format'],
            'title': self.title,
            'organization': self.meta.get('org'),
            'description': self.description,
        })
        context.update(self.meta)

        if self.template:
            context.update(template=self.template)

        if self.filename:
            context.update(filename=self.filename)

        return context
