from commons import UsageError, json
import restclient

import logging
logger = logging.getLogger(__name__)


def _encode_options(options):
    headers = {'Accept': 'application/json;'}
    for item in options:
        headers['Accept'] += '; %s=%s' % (item, json.dumps(json.dumps(options[item])))
    return headers


class ExtAPI(restclient.RESTBase):
    """
    Note: In all the API calls:
    - a 'project' arg variable is the project id
    - a 'task' arg variable is in the format <project_id>-<task_id>
        e.g. '127-T106'
    """

    HIGH_CONFIDENCE_STATUSES = ['pass', 'fail']

    def __init__(self, config):
        extra_conf_opts = [('sde_api_token', 'SDE API Token', '')]
        super(ExtAPI, self).__init__('sde', 'SD Elements', config, 'api/v2', extra_conf_opts)
        self.connected = False

    def post_conf_init(self):
        if self._get_conf('api_token'):
            if '@' not in self._get_conf('api_token'):
                raise UsageError('Unable to process API Token')
            self.config['sde_user'] = None
            self.config['sde_pass'], self.config['sde_server'] = (
                self.config['sde_api_token'].split('@', 1))
            self.set_auth_mode('api_token')

        super(ExtAPI, self).post_conf_init()

    def connect(self):
        # Simple check
        result = self.get_users(myself_only=True)
        self.connected = True
        return bool(result)

    def get_users(self, myself_only=False):
        """
        Gets account information
        """
        end_point = 'users/'
        if myself_only:
            end_point += 'me/'
        return self.call_api(end_point)

    def create_user(self, email, first_name, last_name, role=None, is_active=True, groups=None):
        """
        Create a user
        """
        args = {
            'email': email,
            'first_name': first_name,
            'last_name': last_name,
            'is_active': is_active,
        }

        if role:
            args['role'] = role

        if groups:
            args['groups'] = groups

        return self.call_api('users/', self.URLRequest.POST, args=args)

    def update_user(self, id, **filters):
        """
        Update a user
        """
        args = filters
        end_point = 'users/%d/' % (id)
        return self.call_api(end_point, self.URLRequest.PUT, args=args)

    def get_groups(self):

        response = self.call_api('groups/')
        return response['results']

    def get_global_roles(self):
        response = self.call_api('global-roles/')
        return response['results']

    def get_business_units(self, options={}, **filters):
        """
        Gets all business units accessible to user

        Available Filters:
            name -> business unit name to be searched for
        """
        response = self.call_api('business-units/', args=filters,
                call_headers=_encode_options(options))
        return response['results']

    def get_applications(self, options={}, **filters):
        """
        Gets all applications accessible to user

        Available Filters:
            name -> application name to be searched for
            business_unit -> business unit to search
        """
        response = self.call_api('applications/', args=filters,
                call_headers=_encode_options(options))
        return response['results']

    def create_application(self, name, business_unit):
        args = {
            'business_unit': business_unit,
            'name': name
        }
        return self.call_api('applications/', self.URLRequest.POST, args=args)

    def get_projects(self, application, options={}, **filters):
        """
        Gets all projects for parameter application

        Available Filters:
            name -> project name to be searched for
        """
        args = {
            'application': application
        }
        args.update(filters)
        response = self.call_api('projects/', args=args,
                call_headers=_encode_options(options))
        return response['results']

    def get_tasks(self, project_id, args=None, **filters):
        """
        Get all tasks for a project indicated by the ID of the project
        """
        args.update({'include': 'tags,how_tos,problem,related', 'expand': 'status,phase'})
        args.update(filters)
        response = self.call_api('projects/%d/tasks/' % project_id, args=args)
        return response['results']

    def get_task(self, project_id, task_id, options={}, **filters):
        """
        Get an individual task with parameter task id
        <task> = <project_id>-<task_id> (e.g. 127-T21)
        """
        filters.update({'include': 'tags,how_tos,problem,related', 'expand': 'status,phase'})
        end_point = 'projects/%d/tasks/%s/' % (project_id, task_id)
        return self.call_api(end_point, args=filters, call_headers=_encode_options(options))

    def add_task_note(self, project_id, task_id, text):
        note = {
            'text': text,
        }
        return self.call_api('projects/%d/tasks/%s/notes/' % (project_id, task_id), self.URLRequest.POST, args=note)

    def update_task_note(self, project_id, task_id, note_id, text):
        note = {
            'text': text,
        }
        return self.call_api('projects/%d/tasks/%s/notes/%d/' % (project_id, task_id, note_id), self.URLRequest.PUT, args=note)

    def get_task_notes(self, project_id, task_id, options={}, **filters):
        return self.call_api('projects/%d/tasks/%s/notes/' % (project_id, task_id), call_headers=_encode_options(options))

    def _get_confidence(self, verification_status):
        if verification_status in self.HIGH_CONFIDENCE_STATUSES:
            return 'high'
        else:
            return 'low'

    def add_manual_analysis_note(self, task, verification_status, finding_ref):
        note = {
            'task': task,
            'confidence': self._get_confidence(verification_status),
            'findings': [],
            'behaviour': 'replace',
            'status': verification_status,
            'finding_ref': finding_ref
        }
        return self.call_api('projects/%d/tasks/%s/analysis-notes/', self.URLRequest.POST, args=note)

    def add_analysis_note(self, project_id, task_id, analysis_session, confidence, findings, behaviour, task_status_mapping):
        note = {
            'analysis_session': analysis_session,
            'confidence': confidence,
            'findings': findings,
            'behaviour': behaviour
        }
        if task_status_mapping:
            note['task_status_mapping'] = task_status_mapping
        return self.call_api('projects/%d/tasks/%s/analysis-notes/' % (project_id, task_id), self.URLRequest.POST, args=note)

    def get_analysis_notes(self, project_id, task_id, options={}, **filters):
        return self.call_api('projects/%d/tasks/%s/analysis-notes/' % (project_id, task_id), call_headers=_encode_options(options))

    def create_analysis_session(self, project_id, analysis_ref, analysis_type):
        project_analysis = {
            'analysis_ref': analysis_ref,
            'analysis_type': analysis_type
        }
        return self.call_api('projects/%d/analysis-sessions/' % project_id, self.URLRequest.POST, args=project_analysis)

    def update_task_status(self, project_id, task_id, status):
        """
        Update the task status. The task ID should include the project number
        Returns the 'status' field of the result
        """
        #TODO: regular expression on task and status for validation
        result = self.call_api('projects/%d/tasks/%s/' % (project_id, task_id), self.URLRequest.PATCH, args={'status': status})
        return result['status']

    def get_taskstatuses(self, options={}, **filters):
        """
        Get all statuses for an organization
        """
        response = self.call_api('task-statuses/', args=filters, call_headers=_encode_options(options))
        return response['results']

    def get_phases(self, options={}, **filters):
        """
        Get all phases for an organization
        """
        response = self.call_api('phases/', args=filters, call_headers=_encode_options(options))
        return response['results']

    def get_integration_connections(self, connector_type):
        """
        Get project integration job details
        """
        response = self.call_api('connections/%s/' % connector_type)
        return response['results']

    def update_job_status(self, connection_id, status, msg, connector_type):
        job_status = {
            'connection': connection_id,
            'result_message': msg,
            'succeeded': status,
            'automatic': True
        }
        return self.call_api('jobs/%s/' % connector_type, self.URLRequest.POST, args=job_status)

APIBase = ExtAPI
