import re
import random

from sdetools.sdelib.testlib.response_generator import ResponseGenerator, RESPONSE_HEADERS


class SdeResponseGenerator(ResponseGenerator):
    def __init__(self, config, test_dir=None):
        self.analysis_tools = ['appscan', 'appscan_enterprise', 'veracode', 'fortify', 'webinspect', 'checkmarx']
        self.default_project_id = 1296
        self.default_app_id = 874
        self.default_bu_id = 1
        self.bu_name = config['sde_businessunit']
        self.app_name = config['sde_application']
        self.project_name = config['sde_project']
        resource_templates = ['application.json', 'project.json', 'task.json', 'text_note.json', 'analysis_note.json',
                              'analysis_session.json', 'phases.json', 'taskstatuses.json', 'businessunit.json']
        rest_api_targets = {
            '/api/v2/business-units/': 'call_businessunits',
            '/api/v2/applications/': 'call_applications',
            '/api/v2/projects/[^/]*$': 'get_projects',
            '/api/v2/projects/%s/tasks/[^/]+/[^/]*$' % self.default_project_id : 'call_task',
            '/api/v2/projects/%s/tasks/[^/]*$' % self.default_project_id : 'get_tasks',
            '/api/v2/projects/%s/tasks/[^/]+/notes/' % self.default_project_id: 'call_task_notes',
            '/api/v2/projects/%s/analysis-sessions/' % self.default_project_id: 'add_analysis_session', 
            '/api/v2/projects/%s/tasks/[^/]+/analysis-notes/' % self.default_project_id : 'call_task_analysis_note',
            '/api/v2/task-statuses/': 'get_taskstatuses',
            '/api/v2/phases/': 'get_phases',
            '/api/v2/users/me/$': 'get_users_me',
            '/api/v2/users/$': 'get_users'
        }

        super(SdeResponseGenerator, self).__init__(rest_api_targets, resource_templates, test_dir)

    def init_with_resources(self):
        app_data = {'id': self.default_app_id, 'name': self.app_name, 'business_unit': self.default_bu_id}
        self.generator_add_resource('application', self.default_app_id, app_data)
        project_data = {'name': self.project_name, 'id': self.default_project_id, 'application': self.default_app_id}
        bu_data = {'id': self.default_bu_id, 'name': self.bu_name}
        self.generator_add_resource('businessunit', self.default_bu_id, bu_data)
        self.generator_add_resource('project', self.default_project_id, project_data)
        self.generator_add_resource('task', '40', self.get_json_from_file('T40'))
        self.generator_add_resource('task', '36', self.get_json_from_file('T36'))
        self.generator_add_resource('task', '38', self.get_json_from_file('T38'))
        self.generator_add_resource('phases', '', self.get_json_from_file('phases'))
        self.generator_add_resource('taskstatuses', '', self.get_json_from_file('taskstatuses'))

    def generate_sde_task(self, task_number=None, project_id=None, status=None, priority=7, phase='requirements',
                          tags=None, verification_status=None):
        if task_number is None:
            task_number = '%d' % random.randint(50, 999999999)
        if project_id is None:
            project_id = self.default_project_id
        if status is None:
            status = 'TODO'
        if tags is None:
            tags = []

        sde_task = {
            "id": '%d-T%s' % (project_id, task_number),
            "task_id": 'T%s' % task_number,
            "url": "http://example.com/bunits/new-business-unit/...",
            "title": "Task Title",
            "text": "Insecure forgotten password and password reset...",
            "priority": priority,
            "phase": {
                "id": "X2",
                "name": "Requirements",
                "slug": phase,
                "description": "Application security requirements",
                "tip": "One-time tasks..."
            },
            "ad_hoc": False,
            "relevant": True,
            "accepted": True,
            "assigned_to": [],
            "updated": self.get_current_timestamp(),
            "library_task_created": self.get_current_timestamp(),
            "library_task_updated": self.get_current_timestamp(),
            "verification_status": verification_status,
            "status":  {
                "id": "TS2",
                "meaning": status,
                "icon": "clock-o",
                "name": "Incomplete",
                "slug": status
            },
            "note_count": 0,
            "artifact_proxy": "proxy",
            "tags": tags,
            "related": ["1-T31", "1-T32", "1-T34", "1-T98"],
            "problem": {
                "id": "P526",
                "title": "P526: Weak Password Recovery Mechanism...",
                "text": "It is common for an application...",
                "cwe": [
                    {
                        "url": "http://example.com/640",
                        "id": 640,
                        "title": "Weak Password Recovery..."
                    }
                ]
            },
            "how_tos": [
                {
                    "id": "I131",
                    "title": "I131: Manually with browser",
                    "slug": "test-account-lockout-manually-browser",
                    "url": "http://a7069ccda519b00c4/....",
                    "text": "1. Open your web browser ..."
                }
            ]

        }
        self.generator_add_resource('task', task_number, sde_task)

        return self.generate_resource_from_template('task', sde_task)

    """
       Response functions 
    """
    def get_phases(self, target, flag, data, method):
        if flag:
            self.raise_error('401')

        if method == 'GET':
            params = self.get_url_parameters(target)
            if params:
                phases = self.generator_get_filtered_resource('phases', params)
            else:
                phases = self.generator_get_all_resource('phases')
            return RESPONSE_HEADERS, phases[0]
        else:
            self.raise_error('400')

    def get_users_me(self, target, flag, data, method):
        return self.get_users(target, flag, data, method, myself_only=True)

    def get_users(self, target, flag, data, method, myself_only=False):
        if flag:
            self.raise_error('401')

        if method == 'GET':
            user = {
                "id": 682,
                "email": "frank@sdelements.com",
                "first_name": "Frank",
                "last_name": "Developer",
                "role": {
                    "id": "UR2",
                    "name": "Project Lead"
                },
                "last_login": "2014-12-15T20:10:51.900775Z",
                "date_joined": "2014-04-16T19:43:47.883319Z",
                "is_active": True,
                "is_superuser": False
            }

            if myself_only:
                return RESPONSE_HEADERS, user
            else:
                return RESPONSE_HEADERS, {
                    'results': [ user ]
                }
        else:
            self.raise_error('400')

    def get_taskstatuses(self, target, flag, data, method):
        if flag:
             self.raise_error('401')

        if method == 'GET':
            params = self.get_url_parameters(target)
            if params:
                statuses = self.generator_get_filtered_resource('taskstatuses', params)
            else:
                statuses = self.generator_get_all_resource('taskstatuses')
            return RESPONSE_HEADERS, statuses[0]
        else:
            self.raise_error('400')

    def call_businessunits(self, target, flag, data, method):
        if flag:
            self.raise_error('401')

        if method == 'GET':
            params = self.get_url_parameters(target)
            if params:
                businessunits = self.generator_get_filtered_resource('businessunit', params)
            else:
                businessunits = self.generator_get_all_resource('businessunit')

            return RESPONSE_HEADERS, {'results': businessunits}
        else:
            self.raise_error('400')

    def call_applications(self, target, flag, data, method):
        if flag:
            self.raise_error('401')

        if method == 'GET':
            params = self.get_url_parameters(target)
            if params:
                applications = self.generator_get_filtered_resource('application', params)
            else:
                applications = self.generator_get_all_resource('application')

            return RESPONSE_HEADERS, {'results': applications}
        elif method == 'POST':
            if data:
                self.generator_add_resource('application', resource_data=data)

                return RESPONSE_HEADERS, ''
        self.raise_error('405')

    def get_projects(self, target, flag, data, method):
        if flag:
            self.raise_error('401')

        projects = self.generator_get_filtered_resource('project', self.get_url_parameters(target))
        return RESPONSE_HEADERS, {'results': projects}

    def get_tasks(self, target, flag, data, method):
        if flag:
            self.raise_error('401')

        if method == 'GET':
            return RESPONSE_HEADERS, {'results': self.generator_get_all_resource('task')}
        else:
            self.raise_error('405', {'error': 'Method Not Allowed'})

    def call_task(self, target, flag, data, method):
        if flag:
            self.raise_error('401')

        task_id = target.split('/')[-2].split('-')[1]
        task_number = re.sub('[^\d]+', '', task_id)

        if not self.generator_resource_exists('task', task_number):
            self.raise_error('404', {'error': 'Not Found'})
        if method == 'GET':
            # We will return the task at the end
            pass
        elif method == 'PUT' or method == 'POST':
            if data:
                self.generator_update_resource('task', task_number, data)
            else:
                self.raise_error('400', {'error': 'Missing data param'})
        elif method == 'PATCH':
            if data:
                # We are patching status with slug/meaning
                # When we change to patching using status.id this can be removed
                if 'status' in data:
                    status = data['status']
                    data['status'] = {
                        "id": "TS2",
                        "meaning": status,
                        "icon": "clock-o",
                        "name": "Incomplete",
                        "slug": status
                    }
                
                self.generator_update_resource('task', task_number, data)
            else:
                self.raise_error('400', {'error': 'Missing data param'})
        else:
            self.raise_error('400', {'error': 'Bad Request'})

        return RESPONSE_HEADERS, self.generator_get_resource('task', task_number)

    def call_task_notes(self, target, flag, data, method):
        if flag:
            self.raise_error('401')

        if method == 'GET':
            return RESPONSE_HEADERS, self.generator_get_filtered_resource('text_note', data)
        elif method == 'POST':
            self.generator_add_resource('text_note', resource_data=data)
            return RESPONSE_HEADERS, self.generate_resource_from_template('text_note', data)

    def add_analysis_session(self, target, flag, data, method):
        if flag:
             self.raise_error('401')

        if method == 'GET':
            return RESPONSE_HEADERS, self.generator_get_filtered_resource('analysis_session', data)
        elif method == 'POST':
            if not self.is_data_valid(data, ['analysis_ref']) or not data['analysis_ref']:
                self.raise_error(400)

            if self.is_data_valid(data, ['analysis_type']) and data['analysis_type'] in self.analysis_tools:
                self.generator_add_resource('analysis_session', resource_data=data)
                return RESPONSE_HEADERS, self.generate_resource_from_template('analysis_session', data)

        self.raise_error('500')

    def call_task_analysis_note(self, target, flag, data, method):
        if flag:
            self.raise_error('401')
        task_id = target.split('/')[-3].split('-')[1]
        task_number = re.sub('[^\d]+', '', task_id)

        if method == 'GET':
            data = {}
            data['task_id'] = task_number

            return RESPONSE_HEADERS, self.generator_get_filtered_resource('analysis_note', data)
        elif method == 'POST':
            data['task_id'] = task_number

            if self.is_data_valid(data, ['analysis_session', 'confidence', 'findings']):
                if not data['findings']:
                    data['status'] = 'partial'
                else:
                    data['status'] = 'failed'
            else:
                self.raise_error('400', {'error': 'Missing parameter'})

            self.generator_add_resource('analysis_note', resource_data=data)

            return RESPONSE_HEADERS, self.generate_resource_from_template('analysis_note', data)
        else:
            self.raise_error('400', {'error': 'Invalid request'})
